#include <stdio.h>
#include <pthread.h>
/* #include <GL/glut.h> */
#include <signal.h>
#include <errno.h>  /*  pthread error codes */
#include <sys/time.h>

#include "grind.h"

void getrealcoord(int, int, double*, double*, double*);

/*  thread_id for the graphics thread */
pthread_t gl_tid = 0;

/*  all mouse related variables */
pthread_mutex_t mouse_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t mouse_cond = PTHREAD_COND_INITIALIZER;
boolean mouse_flag =false;
/* the common data */
static double mouse_x = 0.;
static double mouse_y = 0.;
static double mouse_z = 0.;

int readmousecoord(double *nx, double *ny, double *nz) {
  /* called from grind to indicate that grind wants to receive 
     mouse-click coordinates */

  if (!gl_tid)   /* if graphics thread is not running return */
    return 0;

  /*  lock mouse mutex */
  pthread_mutex_lock(&mouse_lock);
  {
    /*  set flag */
    mouse_flag = true;
    /*  wait untill graphics thread has set the data */
    while(mouse_flag)
      pthread_cond_wait(&mouse_cond, &mouse_lock);
    /*  copy relevant data to parameter variables */
    *nx = mouse_x;
    *ny = mouse_y;
    *nz = mouse_z;
  }
  /*  unlock mouse mutex  */
  pthread_mutex_unlock(&mouse_lock);
  return 1;
}

int writemousecoord(int b, int s, int x, int y) {
  /* called from graphics thread to check whether mouse coordinates are wanted
     if so, write them to the common data */

  double wx, wy, wz;  /* returned world x, y, z coords */

  /*  lock mouse mutex before writing the data */
  pthread_mutex_lock(&mouse_lock); 
  {
    /* check whether the grind thread wants mouse coordinates */
    if (mouse_flag) {
      getrealcoord(x, y, &wx, &wy, &wz);
      /* set global data accordingly */
      mouse_x = wx;
      mouse_y = wy;
      mouse_z = wz;
      /* change the flag to indicate we're done */
      mouse_flag = false;
      /* and signal the waiting thread */
      pthread_cond_signal(&mouse_cond);
    }
  }
  /*  and unlock mutex */
  pthread_mutex_unlock(&mouse_lock);
  return 1;
}

/* mutex used for appending and checking the graphics stack  */
pthread_mutex_t stack_lock = PTHREAD_MUTEX_INITIALIZER;

int append2stack(_grcall **last_grcall, _grcall *tmp_grcall) {
  /* called by grind to append a single graphics primitivum 
     to the end of the stack */
  
  /* no need to check whether the other thread exists */

  /* lock stack mutex to block graphics thread to read common data */
  pthread_mutex_lock(&stack_lock);
  /* append the graphics primitivum and adjust grind's last-pointer */
  (*last_grcall)->next = tmp_grcall;
  (*last_grcall) = tmp_grcall;
  /* unlock the stack */
  pthread_mutex_unlock(&stack_lock);
  return 1;
}

int check4append(_grcall **tmp, _grcall **last_of_stack) {
  /* called by graphics thread to check whether grind has appended 
     elements to the stack 
     if new primitiva are found tmp will point to the first new primitivum
     and last_of_stack (gl's last pointer) will point to the new last */
  boolean redraw = false; /* indicate we need to draw new primitiva */
  
  /*  lock stack mutex to see if new primitiva are appended to the stack */
  pthread_mutex_lock(&stack_lock);
  if ((*last_of_stack)->next) {
    redraw = true; /* indiccate we need to draw the new primitiva */
    *tmp = (*last_of_stack)->next; /* point tmp to 1st new primitivum */
    while ((*last_of_stack)->next) /* find the last element of the stack */
      (*last_of_stack) = (*last_of_stack)->next;
  }
  /* all done; unlock mutex */
  pthread_mutex_unlock(&stack_lock);
  /* return whether new elements need to be drawn */
  return redraw;
}

/* mutexes and common data regarding erasing the graphics stack */
pthread_mutex_t erase_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t erase_cond = PTHREAD_COND_INITIALIZER;
/* defines for the mutual info-flag */
#define no_waits 0
#define grind_waits 1
#define gl_waits 2
/* flag to signal the state of stack-erasing */
static int erase_flag = no_waits;

int Signal2ResetStack() {
  /* called by grind thread to indicate that the graphics stack will be erased;
     wait for the graphics stack to respond, then erase, then signal grpahics
     stack that we're done */

  if (!gl_tid)   /* if graphics thread is not running return */
    return 0;

  /* lock erase mutex */
  pthread_mutex_lock(&erase_lock);
  /* set info flag indicating grind is waiting for graphics thread 
     to shake hands */
  erase_flag = grind_waits;
  while (grind_waits == erase_flag)
    /* wait untill other thread responds */
    pthread_cond_wait(&erase_cond, &erase_lock);
  /* we're holding the graphics' hand; at this point erase the stack */
  eraseGRstack();
  /*  signal we're done */
  erase_flag = no_waits;
  pthread_cond_signal(&erase_cond);
  /*  unlock mutex */
  pthread_mutex_unlock(&erase_lock);
  return 1;
}

int checkErase(_grcall *first, _grcall **last) {
  /* called by graphics thread to check whether the graphics stack 
     will be erased
     if so, initiate hand-shaking procedure and 
     reset the graphics-thread owned pointers to the graphics stack */
  boolean erase = false; /* indicate we need to clear the window */

  /*  lock stack mutex to see if there a flag set */
  pthread_mutex_lock(&erase_lock);
  /*  we locked the mutex; now check the info flag */
  if (grind_waits == erase_flag) {
    /* stack will be erased; we will reset the 'last' pointer to 'first' */
    *last = first;
    /* and set flag informing grind we're gonna wait for it to respond */
    erase_flag = gl_waits;
    /* and send the signal */
    pthread_cond_signal(&erase_cond);
    /* we're shacking hands; now we wait untill the stack has been erased */
    while (gl_waits == erase_flag)
    /* wait untill other thread responds */
      pthread_cond_wait(&erase_cond, &erase_lock);
    /*  window needs to be redrawn */
    erase = true;
  }
  /*  unlock stack */
  pthread_mutex_unlock(&erase_lock);
  return erase;
}

